unit TZReader;

{
  written by
                     Dennis Passmore
            EMail: Compuserve: 71640,2464
       http://www.Dennis_Passmore@usgroup.com
}

{ This is a completely rewritten version of the TZip component }
{ by original Author:  Pier Carlo Chiodi                       }
{          E-Mail: pc.chiodi@mbox.thunder.it                   }
{ WEB: http://www.geocities.com/SiliconValley/Park/6159/       }

{
       modifications were made to:

           eliminate memory leaks
           improve overall performance
           add error checking and exception handling
           add multi volume backup set compatibility
           add self extracting exe(SFX) compatibility

  Donated to public domain 02/08/1997

  When installing this component in Delphi 1.0 you will need to
  replace the TzReader.dcr file by renaming it to TzReader.d32
  and then renaming the TzReader.d16 file to TzReader.dcr
}

interface

uses
{$ifdef Win32} Windows, {$else} WinTypes, Winprocs, {$endif}
  Forms, SysUtils, Classes, Dialogs;

{$ifndef ver100    not defined in Delphi 1.0 & 2.0 }
type
  EInvalidOperation = class(exception);
{$endif}

type

 {Local_File_Header_Type}
  ZipFile                = packed Record
    Version              : WORD;
    Flag                 : WORD;
    CompressionMethod    : WORD;
    DateTime             : Longint; { Time: word; Date: Word; }
    CRC32                : Longint;
    CompressedSize       : Longint;
    UncompressedSize     : Longint;
    FileNameLength       : WORD;
    ExtraFieldLength     : WORD;
  { variable length}FileName: String;
  { variable length Extra field }
  end;

  Central_File_Header_Type = packed record
    MadeBy_Version       : Word;
    Extract_Version_Reqd : Word;
    Bit_Flag             : Word;
    Compress_Method      : Word;
    Last_Mod_Time_Date   : Longint;
    Crc32                : LongInt;
    Compressed_Size      : LongInt;
    Uncompressed_Size    : LongInt;
    Filename_Length      : Word;
    Extra_Field_Length   : Word;
    File_Comment_Length  : Word;
    Starting_Disk_Num    : Word;
    Internal_Attributes  : Word;
    External_Attributes  : LongInt;
    Local_Header_Offset  : LongInt;
  { variable length Filename }
  { variable length Extra field }
  { variable length File Comment }
  end;

  End_of_Central_Dir_Type = packed record
    End_Dir_Disk_Number   : Word;
    Central_Dir_Start_Disk: Word;
    Entries_This_Disk     : Word;
    Total_Entries         : Word;
    Central_Dir_Size      : LongInt;
    Start_Disk_Offset     : LongInt;
    ZipFile_Comment_Length: Word;
  { variable length Zip Comment }
  end;

type
  PZipFile = ^ZipFile;

type
  TReadFile = procedure (ZipDirRec: ZipFile) of object;

type
  TZipFileReader = class(TComponent)
  private
    vFiles: TList;
    vFileName: String;
    vFileSize: longint;
    FOnReadFile: TReadFile;   { custom event defined above }
    procedure SetFileName(Value: String);
    function  GetFiles(I: Integer): ZipFile;
    function  GetFileName: string;
    function  GetFileSize: longint;
    function  GetFileCount: Longint;
    function  GetFileDateTime(I: Integer): TDateTime;
    procedure vFiles_Clear;
    function  Prompt_for_Diskette(Diskette_Number: word): boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Clear;
    procedure ReadZipDir(Value: String);
    property  Files[I: Integer]: ZipFile read GetFiles;
    property  FilesDateTime[I: Integer]: TdateTime read GetFileDateTime;
  published
    property  OnReadFile: TReadFile read FOnReadFile write FOnReadFile;
    property  FileName: String read GetFileName write SetFileName;
    property  FileSize: longint read GetFileSize;
    property  FileCount: Longint read GetFileCount;
  end;

procedure Register;

implementation

const
{$ifdef ver80    not defined in Delphi 1.0 }
  HFILE_ERROR             =  integer(-1);
{$endif}
  fLFHBlockSize           =  26;
  Error_reading_Zip_file  = 'Error reading Zip file';
  Invalid_ZIP_file_format = 'Invalid ZIP file format!';
  CentralFileHeaderSig    = $02014b50; { 'PK12' }
  LocalFileHeaderSig      = $04034b50; { 'PK34' }
  EndCentralDirSig        = $06054b50; { 'PK56' }
  MultiVolumeDiskSig      = $08074b50; { 'PK78' }


function MaxI(x, y: longint): longint;
begin
  if x > y then result := x else result := y;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TZipFileReader]);
end;

constructor TZipFileReader.Create(AOWner:TComponent);
begin
  inherited Create(AOwner);
  vFiles := TList.Create;
end;

destructor TZipFileReader.Destroy;
begin
  vFiles_Clear;   { no memory leaks in here }
  vFiles.Free;
  inherited Destroy;
end;

procedure TZipFileReader.vFiles_Clear;
var
  i: integer;
begin
  while (vFiles.Count > 0) do
  begin
    i := vFiles.Count-1;
    if assigned(vFiles[i]) then
      dispose(PZipFile(vFiles[i]));
    vFiles.delete(i);
  end;
end;

procedure TZipFileReader.Clear;
begin
  vFileName := '';
  vFileSize := -1;
  vFiles_Clear;
end;

function TZipFileReader.GetFileDateTime(I:Integer): TDateTime;
begin
  if (I>=0) and (I<vFiles.count) then
    Result := FileDateToDateTime(ZipFile(vFiles[I]^).DateTime)
  else
    raise EListError.Create('Invalid File number');
end;

function TZipFileReader.GetFiles(I:Integer):ZipFile;
begin
  if (I>=0) and (I<vFiles.count) then
    Result := ZipFile(vFiles[I]^)
  else
    raise EListError.Create('Invalid File number');
end;

function TZipFileReader.GetFileName:string;
begin
  if vFileName<>'' then
    Result := vFileName
  else
    raise EInvalidOperation.create('Filename not Assigned!');
end;

function TZipFileReader.GetFileSize:Longint;
begin
  if vFileName<>'' then
    Result := vFileSize
  else
    raise EInvalidOperation.create('Filename not Assigned!');
end;

function TZipFileReader.GetFileCount:Longint;
begin
  if vFileName<>'' then
    Result := vFiles.Count
  else
    raise EInvalidOperation.create('Filename not Assigned!');
end;

function TZipFileReader.Prompt_for_Diskette(Diskette_Number: word): boolean;
var
  s: string;
begin
  s := 'Please insert diskette #'+IntToStr(Diskette_Number)+#0;
  Result := Application.MessageBox(pchar(@s[1]),
                        'Multi Volume Backup Detected', MB_OKCANCEL) = IDOK;
end;

function BackSlashAdjusted(const Filename: string): string;
begin
  Result := Filename;
  while (pos('/',Result) > 0) do
    Result[pos('/',Result)] := '\';
end;

procedure TZipFileReader.ReadZipDir(Value:String);
var
  Done: boolean;
  FirstTime: boolean;
  Sig: Longint;
  Res: Longint;
  ZFile: PZipFile;
{$ifdef Win32}
  Name: shortstring;
{$else}
  Name: string;
{$endif}
  CurrentDisk: integer;
  ZipStream: TFileStream;
  CFile: Central_File_Header_Type;
  EFile: End_of_Central_Dir_Type;

  function Read_Local_File_Headers: boolean;
  begin
    repeat
      new(ZFile);
      fillchar(ZFile^,sizeof(ZFile^),0);

      with ZFile^ do     { fill the structure with local header info for entry }
      if (ZipStream.Read(Version, fLFHBlockSize) = fLFHBlockSize)
        and (ZipStream.Read(Name[1], FileNameLength)=FileNameLength) then
          FileName := BackSlashAdjusted(Copy(Name, 1, FileNameLength))
      else
        raise EStreamError.create(Error_reading_Zip_file);

      vFiles.Add(pointer(ZFile));

 { should never happen due to presence of Central and End directory structures }
      if (ZipStream.Position + ZFile^.ExtraFieldLength + ZFile^.CompressedSize)
                                      > (ZipStream.Size - 22) then
        raise EStreamError.create(Error_reading_Zip_file);

      with ZFile^ do
      begin
        if ExtraFieldLength > 0 then
        begin                     { skip over the extra header attribute fields }
           res := (ZipStream.Position + ExtraFieldLength);
           if ZipStream.Seek(ExtraFieldLength, soFromCurrent) <> res then
             raise EStreamError.create(Error_reading_Zip_file);
        end;
                 { skip over the compressed data for the file entry just parsed }
        res := (ZipStream.Position + CompressedSize);
        if  ZipStream.Seek(CompressedSize, soFromCurrent) <> res then
          raise EStreamError.create(Error_reading_Zip_file);
      end;

                       { Call user's pgm to give it the info on file just read }
      if Assigned(FOnReadFile) then
         FOnReadFile(ZFile^);

      Res := ZipStream.Read(Sig, SizeOf(Sig));
      if (Res = HFILE_ERROR) or (Res <> SizeOf(Sig)) then
        raise EStreamError.create(Error_reading_Zip_file);

    until (ZipStream.Position>=ZipStream.Size) or (Sig = CentralFileHeaderSig);
    result := true;
  end;

  function Read_Central_File_Headers: Boolean;
  var
    x, y: longint;
  begin
    result := false;
    repeat         {  needs to be adjusted to handle split Central directories }
      new(ZFile);
      x := ZipStream.Position;
      fillchar(ZFile^,sizeof(ZFile^),0);
      res := ZipStream.Read(CFile, sizeof(CFile));
      if (res = sizeof(CFile)) then
       with CFile do
        if (ZipStream.Read(Name[1], FileName_Length) = FileName_Length) then
          ZFile^.FileName := BackSlashAdjusted(Copy(Name, 1, FileName_Length))
        else raise EStreamError.create(Error_reading_Zip_file)
      else raise EStreamError.create(Error_reading_Zip_file);

      with ZFile^ do
      begin
        Version              := CFile.Extract_Version_Reqd;
        Flag                 := CFile.Bit_Flag;
        CompressionMethod    := CFile.Compress_Method;
        DateTime             := CFile.Last_Mod_Time_Date;
        CRC32                := CFile.CRC32;
        CompressedSize       := CFile.Compressed_Size;
        UncompressedSize     := CFile.Uncompressed_Size;
        FileNameLength       := CFile.Filename_Length;
        ExtraFieldLength     := CFile.Extra_Field_Length;
      end;

      with ZFile^ do
       if ExtraFieldLength > 0 then        { skip over the extra header fields }
        if ((ZipStream.Position + ExtraFieldLength) <= ZipStream.Size) then
          begin
            res := (ZipStream.Position + ExtraFieldLength);
            if ZipStream.Seek(ExtraFieldLength, soFromCurrent) <> res then
              raise EStreamError.create(Error_reading_Zip_file);
          end
        else begin     { probably at the end of a Split Central directory file }
               dispose(zfile);     { if this is the case the Directory will be }
               break;              { repeated completely on the next diskette. }
             end;

      with CFile do
       if File_Comment_Length > 0 then
        if ((ZipStream.Position + File_Comment_Length) <= ZipStream.Size) then
          begin
            res := (ZipStream.Position + File_Comment_Length);
            if ZipStream.Seek(File_Comment_Length, soFromCurrent) <> res then
              raise EStreamError.create(Error_reading_Zip_file);
          end
        else begin
               dispose(zfile);
               break;
             end;

      vFiles.Add(pointer(ZFile));

                       { Call user's pgm to give it the info on file just read }
      if Assigned(FOnReadFile) then
         FOnReadFile(ZFile^);

      y := ZipStream.Position; { an attempt to approximate total zip file size }
      vFileSize := vFileSize + (y-x) + ZFile^.CompressedSize;

      if ((ZipStream.Position + Sizeof(Sig)) <= ZipStream.Size) then
        begin
          Res := ZipStream.Read(Sig, SizeOf(Sig));
          if (Res = HFILE_ERROR) or (Res <> SizeOf(Sig)) then
            raise EStreamError.create(Error_reading_Zip_file);
        end;

    until ((ZipStream.Position + Sizeof(CFile)) >= ZipStream.Size)
                                        or (Sig = EndCentralDirSig);
    if (Sig = EndCentralDirSig) then
      result := true;
  end;

begin
{$ifdef Win32}
  SetLength(Name,255);
{$else}
  Name[0] := #255;
{$endif}
  vFiles_Clear;
  vFileName := value;
  FirstTime := true;
  ZipStream := nil;
  CurrentDisk := 0;
  try
    ZipStream := TFileStream.Create(Value,fmOpenRead);
    vFileSize := ZipStream.Size;
    ZFile := nil;
    Done := false;
    while not(Done) do
    begin
      Res := ZipStream.Read(Sig, SizeOf(Sig));
      if (Res = HFILE_ERROR) or (Res <> SizeOf(Sig)) then
        raise EStreamError.create(Error_reading_Zip_file)

      else if (Sig = LocalFileHeaderSig)                { 'PK34' }
            and FirstTime then
             begin
               Done := Read_Local_File_Headers;
             end

      else if (Sig = MultiVolumeDiskSig) then           { 'PK78' }
             begin                          { 1st disk of multi volume backup }
               ZipStream.Free;
               ZipStream := nil;
               vFileSize := 0;
               if Application.MessageBox('Please insert last diskette of Set',
                       'Multi Volume Backup Detected', MB_OKCANCEL) = IDOK then
                  ZipStream := TFileStream.Create(Value,fmOpenRead)
               else Done := true;
             end

      else if FirstTime then
             begin            { may be last diskette of multi-disk backup set }
               ZipStream.Seek(0,soFromEnd);   { try to find  EndCentralDirSig }
               FirstTime := false;

               repeat
                 ZipStream.Position := MaxI(0,ZipStream.Position-5);
                 Res := ZipStream.Read(Sig, SizeOf(Sig));      {+4}
                 if (Res = HFILE_ERROR) or (Res <> SizeOf(Sig)) then
                   raise EStreamError.create(Error_reading_Zip_file)
               until (ZipStream.Position=4) or (Sig = EndCentralDirSig);

               if (Sig = EndCentralDirSig) then         { 'PK56' }
                 begin
                   with ZipStream do               { End Directory Block Size }
                     vFileSize := Size - Position - SizeOf(Sig);
                   Res := ZipStream.Read(EFile, SizeOf(EFile));
                   if (Res = HFILE_ERROR) or (Res <> SizeOf(EFile)) then
                     raise EStreamError.create(Error_reading_Zip_file);

                   with EFile do
                    if (Central_Dir_Start_Disk <> End_Dir_Disk_Number) then
                     begin
                       ZipStream.Free;
                       ZipStream := nil;
                       CurrentDisk := Central_Dir_Start_Disk + 1;
                       if Prompt_for_Diskette(CurrentDisk) then
                         ZipStream := TFileStream.Create(Value, fmOpenRead)
                       else Done := true;
                     end;

                   if assigned(ZipStream) then
                     ZipStream.Seek(EFile.Start_Disk_Offset, soFromBeginning);
                 end
               else
                 raise EInvalidOperation.create(Invalid_ZIP_file_format);

             end

      else if (Sig = CentralFileHeaderSig) and          { 'PK12' }
            not (FirstTime) then
             begin                   { may be split across multiple diskettes }
               Done := Read_Central_File_Headers;
               if not(Done) then
               begin
                 ZipStream.Free;
                 ZipStream := nil;
                 CurrentDisk := CurrentDisk + 1;
                 if Prompt_for_Diskette(CurrentDisk) then
                   ZipStream := TFileStream.Create(Value,fmOpenRead)
                 else Done := true;
               end
             end

      else raise EInvalidOperation.create(Invalid_ZIP_file_format);

    end;  { end of while do loop }
  finally
    if assigned(ZipStream) then
      ZipStream.Free;
  end;
end;

procedure TZipFileReader.SetFileName(Value:String);
begin
  if FileExists(Value) then
  begin
    vFileName := value;
    if not(csDesigning in ComponentState) then
      ReadZipDir(Value);
  end;
end;

end.

